﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace BeeDevelopment.Raytracer {

	/// <summary>
	/// Defines a bitmapped texture.
	/// </summary>
	public class BitmapTexture : ITexture {

		#region Fields

		/// <summary>
		/// A 2D array of pixel values for the entire texture.
		/// </summary>
		private Colour[,] pixels;

		#endregion		

		#region Properties

		/// <summary>Gets the width of the texture in pixels.</summary>
		public int Width { get; private set; }
		/// <summary>Gets the height of the texture in pixels.</summary>
		public int Height { get; private set; }

		#endregion

		#region Constructor

		/// <summary>
		/// Creates an instance of a <see cref="BitmapTexture"/> from a file.
		/// </summary>
		/// <param name="filename">The name of the image file to load the texture from.</param>
		public BitmapTexture(string filename) {

			using (Bitmap B = new Bitmap(filename)) {

				this.Width = B.Width;
				this.Height = B.Height;

				var Locked = B.LockBits(new Rectangle(0, 0, this.Width, this.Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
				try {
					var SourceData = new int[Locked.Stride / 4 * Locked.Height];
					Marshal.Copy(Locked.Scan0, SourceData, 0, SourceData.Length);
					pixels = new Colour[this.Width, this.Height];

					for (int y = 0; y < Locked.Height; ++y) {
						int SourcePtr = y * Locked.Stride / 4;
						for (int x = 0; x < Locked.Width; ++x) {
							this.pixels[x, y] = new Colour(SourceData[SourcePtr + x]);
						}
					}

				} finally {
					B.UnlockBits(Locked);
				}
			}
		}
		
		#endregion

		#region Methods

		/// <summary>
		/// Gets the pixel colour from the texture at a specified point.
		/// </summary>
		/// <param name="coordinate">The coordinate to retrieve the colour from.</param>
		/// <returns>The colour from the specified coordinate on the texture.</returns>
		public Colour GetColour(Vector2 coordinate) {

			coordinate.X *= this.Width;
			coordinate.Y *= this.Height;

			Vector2 IntOffset = new Vector2(Math.Floor(coordinate.X), Math.Floor(coordinate.Y));
			Vector2 FractionalOffset = coordinate - IntOffset;

			return Colour.Blend(
				Colour.Blend(this.GetColour((int)IntOffset.X + 0, (int)IntOffset.Y + 0), this.GetColour((int)IntOffset.X + 1, (int)IntOffset.Y + 0), FractionalOffset.X),
				Colour.Blend(this.GetColour((int)IntOffset.X + 0, (int)IntOffset.Y + 1), this.GetColour((int)IntOffset.X + 1, (int)IntOffset.Y + 1), FractionalOffset.X),
				FractionalOffset.Y
			);

		}

		/// <summary>
		/// Gets the colour at a specified integer offset.
		/// </summary>
		/// <param name="u">The X coordinate in pixels.</param>
		/// <param name="v">The Y coordinate in pixels.</param>
		/// <returns>The texture colour at the requested point.</returns>
		/// <remarks>You can pass out-of-bound values to this function - they'll be corrected automatically.</remarks>
		private Colour GetColour(int u, int v) {
			if (u < 0) u = this.Width - ((-u) % this.Width);
			if (v < 0) v = this.Height - ((-v) % this.Height);
			return this.pixels[u % this.Width, v % this.Height];
		}

		#endregion
	}
}
