﻿using System;
using System.Collections.Generic;
using System.Text;

namespace BeeDevelopment.Raytracer {
	public struct Colour {

		private double a, r, g, b;

		/// <summary>Gets the alpha component of the colour.</summary>
		public double A { get { return this.a; } set { this.a = value; } }
		/// <summary>Gets the red component of the colour.</summary>
		public double R { get { return this.r; } set { this.r = value; } }
		/// <summary>Gets the green component of the colour.</summary>
		public double G { get { return this.g; } set { this.g = value; } }
		/// <summary>Gets the blue component of the colour.</summary>
		public double B { get { return this.b; } set { this.b = value; } }


		#region Constructors

		public Colour(double a, double r, double g, double b) {
			this.a = a; this.r = r; this.g = g; this.b = b;
		}

		public Colour(double a, Colour baseColour) {
			this.a = a;
			this.r = baseColour.r;
			this.g = baseColour.g;
			this.b = baseColour.b;
		}

		/// <summary>
		/// Creates an instance of a <see cref="Colour"/> from a 32-bit ARGB value.
		/// </summary>
		/// <param name="argb">The 32-bit ARGB value to create the <see cref="Colour"/> instance from.</param>
		public Colour(int argb) {
			this.a = ((argb >> 24) & 0xFF) / 255.0d;
			this.r = ((argb >> 16) & 0xFF) / 255.0d;
			this.g = ((argb >> 8) & 0xFF) / 255.0d;
			this.b = (argb & 0xFF) / 255.0d;
		}

		#endregion

		#region Operators

		public static Colour operator *(Colour colour, double multiplier) {
			return new Colour(colour.a, colour.r * multiplier, colour.g * multiplier, colour.b * multiplier);
		}

		public static Colour operator +(Colour a, Colour b) {
			return new Colour(a.a + b.a, a.r + b.r, a.g + b.g, a.b + b.b);
		}

		public static Colour operator *(Colour a, Colour b) {
			return new Colour(
				a.a * b.a,
				a.r * b.r,
				a.g * b.g,
				a.b * b.b
			);
		}


		#endregion

		#region Predefined Colours

		/// <summary>Gets the colour black.</summary>
		public static Colour Black { get { return new Colour(1d, 0d, 0d, 0d); } }
		/// <summary>Gets the colour red.</summary>
		public static Colour Red { get { return new Colour(1d, 1d, 0d, 0d); } }
		/// <summary>Gets the colour green.</summary>
		public static Colour Green { get { return new Colour(1d, 0d, 1d, 0d); } }
		/// <summary>Gets the colour blue.</summary>
		public static Colour Blue { get { return new Colour(1d, 0d, 0d, 1d); } }
		/// <summary>Gets the colour cyan.</summary>
		public static Colour Cyan { get { return new Colour(1d, 0d, 1d, 1d); } }
		/// <summary>Gets the colour magenta.</summary>
		public static Colour Magenta { get { return new Colour(1d, 1d, 0d, 1d); } }
		/// <summary>Gets the colour yellow.</summary>
		public static Colour Yellow { get { return new Colour(1d, 1d, 1d, 0d); } }
		/// <summary>Gets the colour white.</summary>
		public static Colour White { get { return new Colour(1d, 1d, 1d, 1d); } }
		/// <summary>Gets the colour grey.</summary>
		public static Colour Grey { get { return new Colour(1d, 0.5d, 0.5d, 0.5d); } }
		/// <summary>Gets the colour transparent.</summary>
		public static Colour Transparent { get { return new Colour(0d, 0d, 0d, 0d); } }

		#endregion

		#region Methods

		public int ToArgb() {
			return
				(Math.Min(255, Math.Max(0, (int)(this.a * 255))) << 24) +
				(Math.Min(255, Math.Max(0, (int)(this.r * 255))) << 16) +
				(Math.Min(255, Math.Max(0, (int)(this.g * 255))) << 8) +
				(Math.Min(255, Math.Max(0, (int)(this.b * 255))) << 0);
		}

		private static double Lerp(double a, double b, double amount) {
			return a + (b - a) * amount;
		}

		public static Colour Blend(Colour baseColour, Colour newColour) {
			return new Colour(1f, 
				Lerp(baseColour.r, newColour.r, newColour.a),
				Lerp(baseColour.g, newColour.g, newColour.a),
				Lerp(baseColour.b, newColour.b, newColour.a)
			);
		}

		public static Colour Blend(Colour baseColour, Colour newColour, double amount) {
			return new Colour(
				Lerp(baseColour.a, newColour.a, amount),
				Lerp(baseColour.r, newColour.r, amount),
				Lerp(baseColour.g, newColour.g, amount),
				Lerp(baseColour.b, newColour.b, amount)
			);
		}

		public static Colour PremultiplyAlpha(Colour c) {
			return new Colour(1.0d, c.r * c.a, c.g * c.a, c.b * c.a);
		}

		#endregion
	}
}
