﻿
using System;
namespace BeeDevelopment.Raytracer {
	public struct Matrix {

		#region Properties

		private double m11; public double M11 { get { return this.m11; } set { this.m11 = value; } }
		private double m12; public double M12 { get { return this.m12; } set { this.m12 = value; } }
		private double m13; public double M13 { get { return this.m13; } set { this.m13 = value; } }
		private double m14; public double M14 { get { return this.m14; } set { this.m14 = value; } }

		private double m21; public double M21 { get { return this.m21; } set { this.m21 = value; } }
		private double m22; public double M22 { get { return this.m22; } set { this.m22 = value; } }
		private double m23; public double M23 { get { return this.m23; } set { this.m23 = value; } }
		private double m24; public double M24 { get { return this.m24; } set { this.m24 = value; } }

		private double m31; public double M31 { get { return this.m31; } set { this.m31 = value; } }
		private double m32; public double M32 { get { return this.m32; } set { this.m32 = value; } }
		private double m33; public double M33 { get { return this.m33; } set { this.m33 = value; } }
		private double m34; public double M34 { get { return this.m34; } set { this.m34 = value; } }

		private double m41; public double M41 { get { return this.m41; } set { this.m41 = value; } }
		private double m42; public double M42 { get { return this.m42; } set { this.m42 = value; } }
		private double m43; public double M43 { get { return this.m43; } set { this.m43 = value; } }
		private double m44; public double M44 { get { return this.m44; } set { this.m44 = value; } }

		#endregion

		#region Constructors

		/// <summary>
		/// Gets an instance of the identity matrix.
		/// </summary>
		public static Matrix Identity {
			get { return new Matrix() { m11 = 1, m22 = 1, m33 = 1, m44 = 1 }; }
		}

		#endregion

		#region Rotation

		// Formulae taken from http://mathworld.wolfram.com/RotationMatrix.html

		/// <summary>
		/// Creates a matrix that represents a rotation in the X axis.
		/// </summary>
		/// <param name="angle">The angle of rotation.</param>
		public static Matrix RotationX(double angle) {
			return new Matrix() {
				m11 = 1,
				m22 = Math.Cos(angle), m23 = Math.Sin(angle),
				m32 = -Math.Sin(angle), m33 = Math.Cos(angle),
				m44 = 1,
			};
		}

		/// <summary>
		/// Creates a matrix that represents a rotation in the Y axis.
		/// </summary>
		/// <param name="angle">The angle of rotation.</param>
		public static Matrix RotationY(double angle) {
			return new Matrix() {
				m11 = Math.Cos(angle), m13 = -Math.Sin(angle),
				m22 = 1,
				m31 = Math.Sin(angle), m33 = Math.Cos(angle),
				m44 = 1,
			};
		}

		/// <summary>
		/// Creates a matrix that represents a rotation in the Z axis.
		/// </summary>
		/// <param name="angle">The angle of rotation.</param>
		public static Matrix RotationZ(double angle) {
			return new Matrix() {
				m11 = Math.Cos(angle), m12 = Math.Sin(angle),
				m21 = -Math.Sin(angle), m22 = Math.Cos(angle),
				m33 = 1,
				m44 = 1,
			};
		}

		#endregion

		#region Arithmetic

		/// <summary>
		/// Multiplies two matrices.
		/// </summary>
		/// <param name="a">The first operand of the matrix multiplication.</param>
		/// <param name="b">The second operand of the matix multiplication.</param>
		/// <returns>The multiplied matrix.</returns>
		public static Matrix Multiply(Matrix a, Matrix b) {
			return new Matrix() {
				m11 = a.m11 * b.m11 + a.m12 * b.m21 + a.m13 * b.m31 + a.m14 * b.m41,
				m12 = a.m11 * b.m12 + a.m12 * b.m22 + a.m13 * b.m32 + a.m14 * b.m42,
				m13 = a.m11 * b.m13 + a.m12 * b.m23 + a.m13 * b.m33 + a.m14 * b.m43,
				m14 = a.m11 * b.m14 + a.m12 * b.m24 + a.m13 * b.m34 + a.m14 * b.m44,
				m21 = a.m21 * b.m11 + a.m22 * b.m21 + a.m23 * b.m31 + a.m24 * b.m41,
				m22 = a.m21 * b.m12 + a.m22 * b.m22 + a.m23 * b.m32 + a.m24 * b.m42,
				m23 = a.m21 * b.m13 + a.m22 * b.m23 + a.m23 * b.m33 + a.m24 * b.m43,
				m24 = a.m21 * b.m14 + a.m22 * b.m24 + a.m23 * b.m34 + a.m24 * b.m44,
				m31 = a.m31 * b.m11 + a.m32 * b.m21 + a.m33 * b.m31 + a.m34 * b.m41,
				m32 = a.m31 * b.m12 + a.m32 * b.m22 + a.m33 * b.m32 + a.m34 * b.m42,
				m33 = a.m31 * b.m13 + a.m32 * b.m23 + a.m33 * b.m33 + a.m34 * b.m43,
				m34 = a.m31 * b.m14 + a.m32 * b.m24 + a.m33 * b.m34 + a.m34 * b.m44,
				m41 = a.m41 * b.m11 + a.m42 * b.m21 + a.m43 * b.m31 + a.m44 * b.m41,
				m42 = a.m41 * b.m12 + a.m42 * b.m22 + a.m43 * b.m32 + a.m44 * b.m42,
				m43 = a.m41 * b.m13 + a.m42 * b.m23 + a.m43 * b.m33 + a.m44 * b.m43,
				m44 = a.m41 * b.m14 + a.m42 * b.m24 + a.m43 * b.m34 + a.m44 * b.m44,
			};
		}

		/// <summary>
		/// Multiplies a matrix and a vector.
		/// </summary>
		/// <param name="m">The matrix to multiply the vector by.</param>
		/// <param name="b">The vector to multiply the matrix by.</param>
		/// <returns>The result of the matrix multiplied by the vector.</returns>
		public static Vector3 Multiply(Matrix m, Vector3 v) {
			return new Vector3(
				v.X * m.m11 + v.Y * m.m21 + v.Z * m.m31,
				v.X * m.m12 + v.Y * m.m22 + v.Z * m.m32,
				v.X * m.m13 + v.Y * m.m23 + v.Z * m.m33
			);
		}

		#endregion

		#region Operators

		public static Matrix operator *(Matrix a, Matrix b) { return Matrix.Multiply(a, b); }
		public static Vector3 operator *(Matrix m, Vector3 v) { return Matrix.Multiply(m, v); }
		public static Vector3 operator *(Vector3 v, Matrix m) { return Matrix.Multiply(m, v); }

		#endregion

	}
}
