﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Threading;
using BeeDevelopment.Raytracer.Lights;

namespace BeeDevelopment.Raytracer {
	public class Raytracer {

		/// <summary>
		/// Used to shunt rays through surfaces they've collided with to avoid reflected/refracted rays from striking the same point again.
		/// </summary>
		private const double SelfInteractionSafetyMargin = 0.0001d;

		#region Types

		private class IntermediateCollision : IComparable<IntermediateCollision> {

			/// <summary>
			/// Gets or sets the collision at this point.
			/// </summary>
			public Collision Collision { get; set; }

			/// <summary>
			/// Gets or sets the total unwound length of the ray's path from the camera.
			/// </summary>
			public double Length { get; set; }

			/// <summary>
			/// Gets or sets the object that was hit.
			/// </summary>
			public WorldObject Object { get; set; }

			public int CompareTo(IntermediateCollision other) {
				return this.Length.CompareTo(other.Length);	
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the list of objects that make up the scene.
		/// </summary>
		public List<WorldObject> Objects { get; private set; }

		/// <summary>
		/// Gets the list of lights that are illuminating the scene.
		/// </summary>
		public List<Light> Lights { get; private set; }

		/// <summary>
		/// Gets or sets the position of the camera.
		/// </summary>
		public Vector3 CameraOrigin { get; set; }

		/// <summary>
		/// Gets or sets the camera's field of view.
		/// </summary>
		public double FieldOfView { get; set; }

		/// <summary>
		/// Gets or sets whether to use linear projection or not.
		/// </summary>
		public bool LinearProjection { get; set; }

		/// <summary>
		/// Gets or sets the rotation of the camera.
		/// </summary>
		public Vector2 CameraRotation { get; set; }

		/// <summary>
		/// Set to true to use 4x supersampling.
		/// </summary>
		public bool EnableAntialiasing { get; set; }

		#endregion

		#region Constructor

		/// <summary>
		/// Creates an instance of the <see cref="Raytracer"/>.
		/// </summary>
		public Raytracer() {
			this.Lights = new List<Light>();
			this.Objects = new List<WorldObject>();
			this.FieldOfView = Math.PI / 2.0d;
			this.LinearProjection = true;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Raytraces the scene.
		/// </summary>
		/// <param name="output">The array of integers to write the output to.</param>
		/// <param name="subRegion">The region to render to.</param>
		/// <param name="height">The stride width of the output array.</param>
		public void Raytrace(int[] output, Rectangle subRegion, Size outputSize, int outputStride) {

			PointF[] SampleOffsets;

			if (this.EnableAntialiasing) {
				var SampleA = 0.3f;
				var SampleB = 0.1f;
				SampleOffsets = new[] { 
					new PointF(0 + SampleA, 0 + SampleB),
					new PointF(1 - SampleA, 0 + SampleB),
					new PointF(0 + SampleA, 1 - SampleB),
					new PointF(1 - SampleA, 1 - SampleB),
				};
			} else {
				SampleOffsets = new[] { new PointF(0.5f, 0.5f) };
			}
			
			double AspectRatio = (double)outputSize.Width / (double)outputSize.Height;
			double TotalAngleX = this.FieldOfView;
			double TotalAngleY = this.FieldOfView / AspectRatio;

			var RotationMatrix = Matrix.RotationX(-CameraRotation.Y) * Matrix.RotationY(CameraRotation.X);

			for (int y = subRegion.Top; y < subRegion.Bottom; ++y) {
				for (int x = subRegion.Left; x < subRegion.Right; x++) {

										
					var Result = Colour.Black;

					foreach (var SampleOffset in SampleOffsets) {

						Vector3 RayDirection = Vector3.Forward;


						Vector2 BaseAngle = new Vector2(
							(((x + SampleOffset.X) - outputSize.Width / 2) / (double)outputSize.Width) * TotalAngleX,
							(((y + SampleOffset.Y) - outputSize.Height / 2) / (double)outputSize.Height) * TotalAngleY
						);

						if (this.LinearProjection) {

							RayDirection.X = BaseAngle.X;
							RayDirection.Y = BaseAngle.Y;
							RayDirection.Z = (TotalAngleX + TotalAngleY) / 2.0d;

							RayDirection *= RotationMatrix;

						} else {
							Vector2 RayAngle = this.CameraRotation + BaseAngle;
							RayDirection.X = Math.Sin(RayAngle.X);
							RayDirection.Y = Math.Sin(RayAngle.Y);
							RayDirection.Z = Math.Cos(RayAngle.X) * Math.Cos(RayAngle.Y);
						}
						
						var CastRay = new Ray(this.CameraOrigin, Vector3.Normalise(RayDirection));

						Result += this.Raytrace(CastRay, 0.0d, 0);	
					}
					Result *= 1.0d / SampleOffsets.Length;
					output[x + y * outputStride] = Result.ToArgb();
				}
				Thread.Sleep(0);
			}
		}

		private Colour Raytrace(Ray ray, double rayOffset, int recursionDepth) {

			if (recursionDepth > 16) return Colour.Black;

			// Get all collisions for the surface.
			var Collisions = new List<IntermediateCollision>();

			foreach (var Object in this.Objects) {
				
				if (Object.Surface != null) {
					var ObjectCollisions = Object.Surface.GetCollision(ray);
					if (ObjectCollisions != null) {
						foreach (var ObjectCollision in ObjectCollisions) {
							Collisions.Add(new IntermediateCollision() {
								Collision = ObjectCollision,
								Length = ObjectCollision.Distance + rayOffset,
								Object = Object,
							});
						}
					}
				}
			}

			// Have we hit anything?
			if (Collisions.Count == 0) return Colour.Black;

			// Get the first collision (and first collision only).
			Collisions.Sort((a, b) => Math.Sign(a.Length - b.Length));
			var C = Collisions[0];

			var StruckColour = C.Object.GetColour(C.Collision.Position);

			// Light the point.
			var LitColour = Colour.Black;
			foreach (var Light in this.Lights) {
				var SubLitColour = StruckColour * Light.Colour;

				// Is it a point light?
				if (Light is PointLight) {

					var PLight = Light as PointLight;
					
					Ray LightToPoint = new Ray(
						PLight.Origin,
						Vector3.Normalise(C.Collision.Position - PLight.Origin)
					);

					LightToPoint.Origin -= SelfInteractionSafetyMargin * LightToPoint.Direction;

					double DistanceToLight = (C.Collision.Position - PLight.Origin).Magnitude;

					SubLitColour *= Vector3.Dot(LightToPoint.Direction, -C.Collision.CollisionNormal);

					// If this is a spotlight, we also need to handle the case where we're outside the beam:
					if (Light is Spotlight) {

						var SLight = Light as Spotlight;
						var AngleFromSpotlight = Vector3.GetAngle(SLight.Direction, LightToPoint.Direction);

						if (AngleFromSpotlight > SLight.InnerAngleLimit) {
							if (AngleFromSpotlight > SLight.OuterAngleLimit) {
								SubLitColour *= 0.0d;
							} else {
								SubLitColour *= 1.0d - (AngleFromSpotlight - SLight.InnerAngleLimit) / (SLight.OuterAngleLimit - SLight.InnerAngleLimit);
							}
						}

					}


					foreach (var ShadingObject in this.Objects) {
						var ShadowCollision = ShadingObject.Surface.GetCollision(LightToPoint);
						if (ShadowCollision != null) {
							foreach (var SC in ShadowCollision) {
								if (SC.Distance < DistanceToLight) {
									// Work out the colour of whatever it is we struck.
									var ShadowColour = ShadingObject.GetColour(SC.Position);
									// Its alpha defines how much to block the shadow ray.
									SubLitColour *= Colour.Blend(Colour.White, ShadowColour, ShadowColour.A);
									SubLitColour *= 1.0d - ShadowColour.A;
								}
							}
						}
					}
				}

				LitColour += SubLitColour;
			}
			

			// Is it translucent?
			if (StruckColour.A < 1.0d) {
				var RefractionRay = new Ray(
					C.Collision.Position,
					Vector3.Refract(ray.Direction, C.Collision.CollisionNormal, C.Collision.Direction == Collision.CollisionDirection.EnteringSolid ? C.Object.Material.RefractiveIndex : 1.0d / C.Object.Material.RefractiveIndex)
				);
				RefractionRay.Origin += SelfInteractionSafetyMargin * RefractionRay.Direction;
				LitColour = Colour.Blend(LitColour, this.Raytrace(RefractionRay, C.Length + SelfInteractionSafetyMargin, recursionDepth + 1), 1.0d - StruckColour.A);
			}

			// Is it reflective?
			if (C.Object.Material.Reflectivity > 0.0d) {
				var ReflectionRay = new Ray(
					C.Collision.Position,
					Vector3.Reflect(ray.Direction, C.Collision.CollisionNormal)
				);
				ReflectionRay.Origin += SelfInteractionSafetyMargin * ReflectionRay.Direction;
				LitColour = Colour.Blend(LitColour, this.Raytrace(ReflectionRay, C.Length + SelfInteractionSafetyMargin, recursionDepth + 1), C.Object.Material.Reflectivity);
			}

			return LitColour;

		}
		#endregion
	}
}
