﻿using System.Collections.Generic;

namespace BeeDevelopment.Raytracer.Surfaces.Csg {
	/// <summary>
	/// Represents the intersection of two surfaces.
	/// </summary>
	/// <remarks>Only collisions that are within surface A and surface B are returned.</remarks>
	public class Intersection : IRayCollidable, ITexturable {

		/// <summary>
		/// Gets the operand surface A.
		/// </summary>
		public IRayCollidable A { get; set; }

		/// <summary>
		/// Gets the operand surface B.
		/// </summary>
		public IRayCollidable B { get; set; }

		public Intersection(IRayCollidable a, IRayCollidable b) {
			this.A = a;
			this.B = b;
		}

		#region IRayCollidable Members

		public Collision[] GetCollision(Ray ray) {
			
			var ACollisions = new List<Collision>(A.GetCollision(ray) ?? new Collision[0]);
			//if (ACollisions.Count == 0) return null;

			var BCollisions = new List<Collision>(B.GetCollision(ray) ?? new Collision[0]);
			//if (BCollisions.Count == 0) return null;

			ACollisions.RemoveAll(c => !B.PointIsInside(c.Position));
			BCollisions.RemoveAll(c => !A.PointIsInside(c.Position));
			
			// Return all collisions.
			var AllCollisions = new List<Collision>();
			AllCollisions.AddRange(ACollisions);
			AllCollisions.AddRange(BCollisions);
			return AllCollisions.ToArray();
		}

		public bool PointIsInside(Vector3 point) {
			return A.PointIsInside(point) && B.PointIsInside(point);
		}

		public Vector3 GetClosestPoint(Vector3 point) {
			Vector3 OnA = this.A.GetClosestPoint(point);
			Vector3 OnB = this.B.GetClosestPoint(point);
			if ((OnA - point).MagnitudeSquared < (OnB - point).MagnitudeSquared) {
				return OnA;
			} else {
				return OnB;
			}
		}

		#endregion

		#region ITexturable Members

		public Vector2 GetTextureCoordinate(Vector3 surfacePoint) {
			
			Vector3 OnA = this.A.GetClosestPoint(surfacePoint);
			Vector3 OnB = this.B.GetClosestPoint(surfacePoint);
			
			if ((OnA - surfacePoint).MagnitudeSquared < (OnB - surfacePoint).MagnitudeSquared) {
				if (this.A is ITexturable) return (this.A as ITexturable).GetTextureCoordinate(surfacePoint);
			} else {
				if (this.B is ITexturable) return (this.B as ITexturable).GetTextureCoordinate(surfacePoint);
			}

			return Vector2.Zero;
		}


		#endregion
	}
}
