﻿using System;
using System.Collections.Generic;
using System.Text;

namespace BeeDevelopment.Raytracer.Surfaces {
	public class Cylinder : IRayCollidable, ITexturable {

		public Vector3 StartPoint { get; set; }
		public Vector3 EndPoint { get; set; }
		public double Radius { get; set; }

		/// <summary>
		/// Gets or sets whether the start of the cylinder is called.
		/// </summary>
		public bool StartCapped { get; set; }

		/// <summary>
		/// Gets or sets whether the start of the cylinder is called.
		/// </summary>
		public bool EndCapped { get; set; }

		public Collision[] GetCollision(Ray ray) {

			double RR = this.Radius * this.Radius;

			var Collisions = new List<Collision>();

			// Start by trying to collide with the caps.
			if (this.StartCapped) {
				var N = Vector3.Normalise(this.StartPoint - this.EndPoint);
				var P = new Plane(N, Vector3.Dot(this.StartPoint, N));
				var C = P.GetCollision(ray);
				if (C != null && C.Length > 0 && (C[0].Position - this.StartPoint).MagnitudeSquared < RR) {
					Collisions.Add(C[0]);
				}
			}

			if (EndCapped) {
				var N = Vector3.Normalise(this.EndPoint - this.StartPoint);
				var P = new Plane(N, Vector3.Dot(this.EndPoint, N));
				var C = P.GetCollision(ray);
				if (C != null && C.Length > 0 && (C[0].Position - this.EndPoint).MagnitudeSquared < RR) {
					Collisions.Add(C[0]);
				}
			}

			Vector3 AB = EndPoint - StartPoint;
			Vector3 AO = ray.Origin - StartPoint;
			Vector3 AOxAB = Vector3.Cross(AO, AB);
			Vector3 VxAB = Vector3.Cross(ray.Direction, AB);
			double ab2 = Vector3.Dot(AB, AB);
			double a = Vector3.Dot(VxAB, VxAB);
			double b = 2 * Vector3.Dot(VxAB, AOxAB);
			double c = Vector3.Dot(AOxAB, AOxAB) - (RR * ab2);

			// at² + bt + c = 0

			double Q = b * b - 4.0d * a * c;
			if (Q >= 0) {

				Q = Math.Sqrt(Q);

				for (int Multiplier = -1; Multiplier <= +1; Multiplier += 2) {

					double T = (-b + Q * Multiplier) / (2.0d * a);

					if (T < 0) {
						continue;
					} else {

						var StruckPoint = ray.Origin + ray.Direction * T;


						double DistanceAlongAxis = Vector3.Dot(StruckPoint - this.StartPoint, this.EndPoint - this.StartPoint);
						double CylinderLengthSquared = (EndPoint - StartPoint).MagnitudeSquared;

						double AmountAlongAxis = DistanceAlongAxis / CylinderLengthSquared;

						if (AmountAlongAxis < 0 || AmountAlongAxis > 1) {
							continue;
						}

						var PointOnAxis = StartPoint + (EndPoint - StartPoint) * AmountAlongAxis;

						var SurfaceNormal = Vector3.Normalise(StruckPoint - PointOnAxis);
						
						Collisions.Add(
							new Collision(
									Vector3.Dot(SurfaceNormal, ray.Direction) > 0 ? Collision.CollisionDirection.LeavingSolid : Collision.CollisionDirection.EnteringSolid, 
									StruckPoint, T, SurfaceNormal
							)
						);
					}
				}
			}

			return Collisions.ToArray();

		}

		public bool PointIsInside(Vector3 point) {

			Vector3 Axis = this.EndPoint - this.StartPoint;

			double AmountAlongAxis = Vector3.Dot(Axis, point - this.StartPoint) / Axis.MagnitudeSquared;

			if (AmountAlongAxis < 0.0d || AmountAlongAxis > 1.0d) return false; // Past the end!

			double DistanceFromAxisSquared = ((AmountAlongAxis * Axis + this.StartPoint) - point).MagnitudeSquared;

			return  DistanceFromAxisSquared < this.Radius * this.Radius;
		}

		public Vector2 GetTextureCoordinate(Vector3 surfacePoint) {

			var Axis = (EndPoint - StartPoint);

			double DistanceAlongAxis = Vector3.Dot(surfacePoint - this.StartPoint, this.EndPoint - this.StartPoint);
			double CylinderLengthSquared = (EndPoint - StartPoint).MagnitudeSquared;
			double AmountAlongAxis = DistanceAlongAxis / CylinderLengthSquared;
			var PointOnAxis = StartPoint + (EndPoint - StartPoint) * AmountAlongAxis;

			// Create a vector perpendicular to the axis:
			var AbsAxis = Vector3.Abs(Axis);
			Vector3 PerpendicularVector;
			if (AbsAxis.X < AbsAxis.Y && AbsAxis.X < AbsAxis.Z) {
				PerpendicularVector = Vector3.Cross(Axis, new Vector3(+1.0d, 0.0d, 0.0d));
			} else if (AbsAxis.Y < AbsAxis.X && AbsAxis.Y < AbsAxis.Z) {
				PerpendicularVector = Vector3.Cross(Axis, new Vector3(0.0d, +1.0d, 0.0d));
			} else {
				PerpendicularVector = Vector3.Cross(Axis, new Vector3(0.0d, 0.0d, +1.0d));
			}

			var AxisToPoint = surfacePoint - PointOnAxis;
			var PerpendicularDotAxis = Vector3.Dot(PerpendicularVector, AxisToPoint);
			double RotationalAngle = Math.Acos(PerpendicularDotAxis / (PerpendicularVector.Magnitude * AxisToPoint.Magnitude)) / (Math.PI * 2.0d);

			if (Vector3.Dot(Vector3.Cross(AxisToPoint, Axis), PerpendicularVector) > 0.0d) RotationalAngle = 1.0d - RotationalAngle;

			return new Vector2(RotationalAngle, AmountAlongAxis);
		}

		public Vector3 GetClosestPoint(Vector3 point) {
			//TODO: Not done (laziness).
			return Vector3.Zero;
		}
	}
}
