﻿
namespace BeeDevelopment.Raytracer.Surfaces {
	/// <summary>Defines a plane.</summary>
	public class Plane : IRayCollidable, ITexturable {

		#region Types

		/// <summary>Describes the side of a plane something is.</summary>
		public enum Side {
			/// <summary>The item is in front of the plane.</summary>
			Front,
			/// <summary>The item is in behind the plane.</summary>
			Back,
			/// <summary>The item lies inside the plane.</summary>
			Inside,
		}

		#endregion

		#region Properties

		private Vector3 normal;
		/// <summary>Gets or sets the plane normal.</summary>
		public Vector3 Normal { get { return this.normal; } set { this.normal = value; } }

		private double offset;
		/// <summary>Gets or sets the offset of the plane from the origin.</summary>
		public double Offset { get { return this.offset; } set { this.offset = value; } }

		private Vector3 textureXAxis;
		/// <summary>
		/// Gets or sets the X axis to align textures to.
		/// </summary>
		public Vector3 TextureXAxis { get { return this.textureXAxis; } set { this.textureXAxis = value; } }

		private Vector3 textureYAxis;
		/// <summary>
		/// Gets or sets the Y axis to align textures to.
		/// </summary>
		public Vector3 TextureYAxis { get { return this.textureYAxis; } set { this.textureYAxis = value; } }

		/// <summary>
		/// Gets or sets the 2D texture offset.
		/// </summary>
		public Vector2 TextureOffset { get; set; }

		/// <summary>
		/// Gets or sets the 3D texture origin.
		/// </summary>
		public Vector3 TextureOrigin { get; set; }

		#endregion

		#region Constructor

		/// <summary>Creates an instance of the plane from a normal and offset value.</summary>
		/// <param name="normal">The plane's normal.</param>
		/// <param name="offset">The offset of the plane along the normal from the origin.</param>
		public Plane(Vector3 normal, double offset) {
			this.normal = normal;
			this.offset = offset;
			this.textureXAxis = Vector3.Right;
			this.textureYAxis = Vector3.Up;
		}

		/// <summary>
		/// Creates an instance of a plane from a normal and a point on the plane.
		/// </summary>
		/// <param name="normal">The surface normal of the plane.</param>
		/// <param name="pointOnPlane"></param>
		public Plane(Vector3 normal, Vector3 pointOnPlane)
			: this(normal, Vector3.Dot(normal, pointOnPlane)) {
		}

		#endregion

		#region Methods

		/// <summary>Get the side of the plane a point lies in.</summary>
		/// <param name="p">The position of the point to check.</param>
		public Side GetSide(Vector3 p) {
			double f = Vector3.Dot(this.Normal, p) - this.Offset;
			if (f < 0) {
				return Side.Front;
			} else if (f > 0) {
				return Side.Back;
			} else {
				return Side.Inside;
			}
		}

		public Collision[] GetCollision(Ray ray) {
			
			Vector3 FacingNormal = -this.Normal;
			double DotNormal = Vector3.Dot(ray.Direction, FacingNormal);

			double Distance = -(Vector3.Dot(ray.Origin, FacingNormal) + this.Offset) / DotNormal;
			if (Distance >= 0 && !double.IsInfinity(Distance)) {
				return new[] { new Collision(DotNormal < 0.0d ? Collision.CollisionDirection.LeavingSolid : Collision.CollisionDirection.EnteringSolid, ray.Origin + ray.Direction * Distance, Distance, this.Normal) };
			} else {
				return null;
			}
		}

		public bool PointIsInside(Vector3 point) {
			return (Vector3.Dot(this.Normal, point) - this.Offset) < 0.0d;
		}

		public Vector2 GetTextureCoordinate(Vector3 surfacePoint) {
			surfacePoint -= this.TextureOrigin;
			return new Vector2(Vector3.Dot(surfacePoint, this.textureXAxis), Vector3.Dot(surfacePoint, this.textureYAxis)) + this.TextureOffset;
		}

		public Vector3 GetClosestPoint(Vector3 point) {
			double Distance = Vector3.Dot(this.normal, point - this.normal * this.offset);
			return point - this.normal * Distance;
		}

		#endregion
	}
}
