﻿
using BeeDevelopment.Raytracer.Surfaces.Csg;
namespace BeeDevelopment.Raytracer.Surfaces {
	public class SurfaceBuilder {

		/// <summary>Builds a cuboid.</summary>
		/// <param name="origin">The centre of the cuboid in 3D space.</param>
		/// <param name="topNormal">The surface normal of the top face.</param>
		/// <param name="frontNormal">The surface normal of the front face.</param>
		/// <param name="width">The width of the cuboid (difference between left and right faces).</param>
		/// <param name="height">The height of the cuboid (difference between top and bottom faces).</param>
		/// <param name="depth">The depth of the cuboid (difference between the front and back faces).</param>
		public static IRayCollidable BuildCuboid(Vector3 origin, Vector3 topNormal, Vector3 frontNormal, double width, double height, double depth) {

			topNormal = Vector3.Normalise(topNormal);
			frontNormal = Vector3.Normalise(frontNormal);
			var rightNormal = Vector3.Cross(topNormal, frontNormal);

			var FrontAndBack = new Intersection(
				new Plane( frontNormal, origin + frontNormal * (depth * 0.5d)) {
					TextureXAxis = rightNormal / width,
					TextureYAxis = -topNormal / height,
					TextureOrigin = origin + (frontNormal * depth * 0.5d) + (-rightNormal * width * 0.5d) + (topNormal * height * 0.5d),
				},
				new Plane(-frontNormal, origin - frontNormal * (depth * 0.5d)) {
					TextureXAxis = -rightNormal / width,
					TextureYAxis = -topNormal / height,
					TextureOrigin = origin + (-frontNormal * depth * 0.5d) + (rightNormal * width * 0.5d) + (topNormal * height * 0.5d),
				}
			);

			var TopAndBottom = new Intersection(
				new Plane(topNormal, origin + topNormal * (height * 0.5d)) {
					TextureXAxis = rightNormal / width,
					TextureYAxis = frontNormal / depth,
					TextureOrigin = origin + (-frontNormal * depth * 0.5d) + (-rightNormal * width * 0.5d) + (topNormal * height * 0.5d),
				},
				new Plane(-topNormal, origin - topNormal * (height * 0.5d)) {
					TextureXAxis = rightNormal / width,
					TextureYAxis = -frontNormal / depth,
					TextureOrigin = origin + (frontNormal * depth * 0.5d) + (-rightNormal * width * 0.5d) + (-topNormal * height * 0.5d),
				}
			);

			var LeftAndRight = new Intersection(
				new Plane(rightNormal, origin + rightNormal * (height * 0.5d)) {
					TextureXAxis = -frontNormal / depth,
					TextureYAxis = -topNormal / height,
					TextureOrigin = origin + (frontNormal * depth * 0.5d) + (rightNormal * width * 0.5d) + (topNormal * height * 0.5d),
				},
				new Plane(-rightNormal, origin - rightNormal * (height * 0.5d)) {
					TextureXAxis = frontNormal / depth,
					TextureYAxis = -topNormal / height,
					TextureOrigin = origin + (-frontNormal * depth * 0.5d) + (-rightNormal * width * 0.5d) + (topNormal * height * 0.5d),
				}
			);

			return new Intersection(FrontAndBack, new Intersection(TopAndBottom, LeftAndRight));

		}

		/// <summary>Builds a cube.</summary>
		/// <param name="origin">The centre of the cuboid in 3D space.</param>
		/// <param name="topNormal">The surface normal of the top face.</param>
		/// <param name="frontNormal">The surface normal of the front face.</param>
		/// <param name="size">The size the cuboid.</param>
		public static IRayCollidable BuildCube(Vector3 origin, Vector3 topNormal, Vector3 frontNormal, double size) {
			return SurfaceBuilder.BuildCuboid(origin, topNormal, frontNormal, size, size, size);
		}

	}
}
