﻿using System.Windows.Forms;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System;
using System.Threading;
using BeeDevelopment.Raytracer.Surfaces;
using BeeDevelopment.Raytracer.Lights;
using BeeDevelopment.Raytracer.Surfaces.Csg;

namespace BeeDevelopment.Raytracer.UI {
	public partial class MainInterface : Form {


		private int[] OutputDiffuse;
		private const int OutputWidth = 640;
		private const int OutputHeight = 480;
		private Bitmap FinalOutputBuffer;

		private Raytracer Tracer;

		private struct TracedCollision {
			private Collision collision;
			public Collision Collision { get { return this.collision; } set { this.collision = value; } }

			private double pathLength;
			public double PathLength { get { return this.pathLength; } set { this.pathLength = value; } }

			private WorldObject hitObject;
			public WorldObject HitObject { get { return this.hitObject; } set { this.hitObject = value; } }

			private int reflectionCount;
			public int ReflectionCount { get { return this.reflectionCount; } set { this.reflectionCount = value; } }

		}		

		public MainInterface() {

			InitializeComponent();
			this.Text = Application.ProductName;

			this.OutputDiffuse = new int[OutputWidth * OutputHeight];
			FinalOutputBuffer = new Bitmap(OutputWidth, OutputHeight);

			this.ClientSize = new Size(OutputWidth, OutputHeight);

			this.Disposed += (sender, e) => this.FinalOutputBuffer.Dispose();

			this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw | ControlStyles.UserPaint, true);

			this.Tracer = new Raytracer() {
				CameraOrigin = new Vector3(9, -5, 0),
				CameraRotation = new Vector2(-0.3, 0.2),
				FieldOfView = Math.PI / 4.0d,
				LinearProjection = true,
				EnableAntialiasing = true,
			};
			
			// Rear wall.
			this.Tracer.Objects.Add(new WorldObject(
				new Plane(Vector3.Backward, -50d) { TextureXAxis = Vector3.Right / 20d, TextureYAxis = Vector3.Down / 20d, TextureOffset = Vector2.Half },
				new Material() { Texture = new BitmapTexture("brick.jpg") }
			));

			// Floor.
			this.Tracer.Objects.Add(new WorldObject(
				new Plane(Vector3.Up, -10d) { TextureXAxis = Vector3.Right / 20d, TextureYAxis = Vector3.Backward / 20d, TextureOrigin = new Vector3(-10,10,50) },
				new Material() { Texture = new BitmapTexture("marble.jpg"), Reflectivity = 0.3 }
			));

			// Ceiling.
			this.Tracer.Objects.Add(new WorldObject(
				new Plane(Vector3.Down, -10d),
				new Material()
			));

			// Left wall.
			this.Tracer.Objects.Add(new WorldObject(
				new Plane(Vector3.Right, -10d),
				new Material()
			));

			// Right wall.
			this.Tracer.Objects.Add(new WorldObject(
				new Plane(Vector3.Left, -10d),
				new Material()
			));


			// Chuck in some objects:
			this.Tracer.Objects.Add(new WorldObject(
				SurfaceBuilder.BuildCube(new Vector3(-5, 7, 40), Vector3.Up, Vector3.Backward * Matrix.RotationY(-0.3), 6),
				new Material() { Colour = Colour.Green, Reflectivity = 0.4 }
			));
			this.Tracer.Objects.Add(new WorldObject(
				SurfaceBuilder.BuildCube(new Vector3(+5, 7, 40), Vector3.Up, Vector3.Backward * Matrix.RotationY(0.3), 6),
				new Material() { Colour = Colour.Blue, Reflectivity = 0.4 }
			));
			this.Tracer.Objects.Add(new WorldObject(
				SurfaceBuilder.BuildCube(new Vector3(0, 1, 40), Vector3.Up, Vector3.Backward, 6),
				new Material() { Colour = Colour.Red, Reflectivity = 0.4 }
			));



			this.Tracer.Lights.Add(new Spotlight(new Vector3(0, -9.5, 25), Vector3.Normalise(new Vector3(0, 1, 1)), 0.3, 0.6));

			this.Tracer.Lights.Add(new PointLight(new Vector3(-3, 1, 0), Colour.Grey * 0.05));
			this.Tracer.Lights.Add(new PointLight(new Vector3(+3, 1, 0), Colour.Grey * 0.05));

			//this.Tracer.Lights.Clear(); this.Tracer.Lights.Add(new Light());		

			// Start raytracing.
			this.Raytrace();
		}

		private void UpdateFinalOutputBuffer() {
			var Locked = this.FinalOutputBuffer.LockBits(new Rectangle(0, 0, OutputWidth, OutputHeight), System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb);
			try {
				Marshal.Copy(this.OutputDiffuse, 0, Locked.Scan0, this.OutputDiffuse.Length);
			} finally {
				this.FinalOutputBuffer.UnlockBits(Locked);
			}
		}

		protected override void OnPaint(PaintEventArgs e) {
			this.UpdateFinalOutputBuffer();
			e.Graphics.DrawImage(this.FinalOutputBuffer, this.ClientRectangle);
		}

		private List<Thread> Threads = new List<Thread>();

		private void Raytrace() {
			int ThreadCount = 3;
			int SubregionHeight = OutputHeight / ThreadCount;
			for (int i = 0; i < ThreadCount; ++i) {
				int SubregionIndex = i;
				var T = new Thread(() => this.Tracer.Raytrace(this.OutputDiffuse, new Rectangle(0, SubregionIndex * SubregionHeight, OutputWidth, SubregionHeight), new Size(OutputWidth, OutputHeight), OutputWidth));
				T.Start();
				Threads.Add(T);
			}
		}

		private void CancelRaytrace() {
			foreach (var T in this.Threads) 
				if (T.IsAlive) T.Abort();
		}

		private void Invalidator_Tick(object sender, EventArgs e) {
			this.Invalidate();
		}

		private void MainInterface_FormClosed(object sender, FormClosedEventArgs e) {
			this.CancelRaytrace();
		}
	}
}

