﻿using System;

namespace BeeDevelopment.Raytracer {

	/// <summary>
	/// Defines a vector with two components (X, Y and Z).
	/// </summary>
	public struct Vector2 {

		private double x, y;

		#region Instance Properties

		/// <summary>Gets or sets the X component of the vector.</summary>
		public double X { get { return this.x; } set { this.x = value; } }
		/// <summary>Gets or sets the Y component of the vector.</summary>
		public double Y { get { return this.y; } set { this.y = value; } }


		/// <summary>Gets or sets the magnitude of the vector.</summary>
		public double Magnitude {
			get {
				return Math.Sqrt(x * x + y * y);
			}
			set {
				var Normalised = Vector2.Normalise(this);
				this.x = Normalised.x * value;
				this.y = Normalised.y * value;
			}
		}

		/// <summary>Gets the magnitude squared of the vector.</summary>
		public double MagnitudeSquared {
			get {
				return x * x + y * y;
			}
		}

		#endregion

		#region Static Properties

		/// <summary>Gets a Vector2 with all components set to zero.</summary>
		public static Vector2 Zero {
			get { return new Vector2(0.0d); }
		}

		/// <summary>Gets a Vector2 with all components set to 0.5.</summary>
		public static Vector2 Half {
			get { return new Vector2(0.5d); }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Creates an instance of a vector with the two components set to specified values.
		/// </summary>
		/// <param name="x">The value of the vector's X component.</param>
		/// <param name="y">The value of the vector's Y component.</param>
		public Vector2(double x, double y) {
			this.x = x; this.y = y;
		}

		/// <summary>
		/// Creates an instance of a vector with the two components set to the same value.
		/// </summary>
		/// <param name="v">The value to set all two components to.</param>
		public Vector2(double v) {
			this.x = v; this.y = v;
		}

		#endregion

		#region General Methods

		/// <summary>Returns a formatted string describing the vector's components.</summary>
		public override string ToString() {
			return string.Format("'{0:N1} {1:N1}'", this.X, this.Y);
		}

		/// <summary>Gets the hash code for this vector.</summary>
		public override int GetHashCode() {
			return this.X.GetHashCode() ^ this.Y.GetHashCode();
		}

		#endregion

		#region Arithmetic Methods

		/// <summary>
		/// Normalises a vector.
		/// </summary>
		/// <param name="v">The vector to normalise.</param>
		/// <returns>A vector of unit length.</returns>
		public static Vector2 Normalise(Vector2 v) {
			double Multiplier = 1f / v.Magnitude;
			return new Vector2() {
				X = v.x * Multiplier,
				Y = v.y * Multiplier,
			};
		}

		/// <summary>Calculates the dot (scalar) product of two vectors.</summary>
		public static double Dot(Vector2 a, Vector2 b) {
			return a.X * b.X + a.Y * b.Y;
		}

		/// <summary>Returns a vector made up of the smallest components of two vectors.</summary>
		public static Vector2 Min(Vector2 a, Vector2 b) {
			return new Vector2(Math.Min(a.X, b.X), Math.Min(a.Y, b.Y));
		}


		/// <summary>Returns a vector made up of the largest components of two vectors.</summary>
		public static Vector2 Max(Vector2 a, Vector2 b) {
			return new Vector2(Math.Max(a.X, b.X), Math.Max(a.Y, b.Y));
		}

		#endregion

		#region Operators

		/// <summary>Adds the components of two vectors together.</summary>
		public static Vector2 operator +(Vector2 a, Vector2 b) { return new Vector2(a.X + b.X, a.Y + b.Y); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector2 operator +(Vector2 a, double b) { return a + new Vector2(b); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector2 operator +(double a, Vector2 b) { return new Vector2(a) + b; }

		/// <summary>Subtracts the components of one vector from another.</summary>
		public static Vector2 operator -(Vector2 a, Vector2 b) { return new Vector2(a.X - b.X, a.Y - b.Y); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector2 operator -(Vector2 a, double b) { return a - new Vector2(b); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector2 operator -(double a, Vector2 b) { return new Vector2(a) - b; }

		/// <summary>Negates the components of a vector.</summary>
		public static Vector2 operator -(Vector2 a) { return new Vector2(-a.X, -a.Y); }

		/// <summary>Multiplies the components of one vector with another.</summary>
		public static Vector2 operator *(Vector2 a, Vector2 b) { return new Vector2(a.X * b.X, a.Y * b.Y); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector2 operator *(Vector2 a, double b) { return a * new Vector2(b); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector2 operator *(double a, Vector2 b) { return new Vector2(a) * b; }

		/// <summary>Divides the components of one vector by the components of another.</summary>
		public static Vector2 operator /(Vector2 a, Vector2 b) { return new Vector2(a.X / b.X, a.Y / b.Y); }

		/// <summary>Divides the components of a vector by a constant value.</summary>
		public static Vector2 operator /(Vector2 a, double b) { return a / new Vector2(b); }

		/// <summary>Divides a vector with all components set to a single value by another vector.</summary>
		public static Vector2 operator /(double a, Vector2 b) { return new Vector2(a) / b; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public static bool operator ==(Vector2 a, Vector2 b) { return a.X == b.X && a.Y == b.Y; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have differing components, false if both are the same.</returns>
		public static bool operator !=(Vector2 a, Vector2 b) { return a.X != b.X || a.Y != b.Y; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public override bool Equals(object obj) {
			if (obj.GetType() != this.GetType()) return false;
			return (Vector2)obj == this;
		}

		#endregion

	}
}
