﻿using System;

namespace BeeDevelopment.Raytracer {

	/// <summary>
	/// Defines a vector with three components (X, Y and Z).
	/// </summary>
	public struct Vector3 {

		private double x, y, z;

		#region Instance Properties

		/// <summary>Gets or sets the X component of the vector.</summary>
		public double X { get { return this.x; } set { this.x = value; } }
		/// <summary>Gets or sets the Y component of the vector.</summary>
		public double Y { get { return this.y; } set { this.y = value; } }
		/// <summary>Gets or sets the Z component of the vector.</summary>
		public double Z { get { return this.z; } set { this.z = value; } }

		/// <summary>Gets or sets the magnitude of the vector.</summary>
		public double Magnitude {
			get {
				return Math.Sqrt(x * x + y * y + z * z);
			}
			set {
				var Normalised = Vector3.Normalise(this);
				this.x = Normalised.x * value;
				this.y = Normalised.y * value;
				this.z = Normalised.z * value;
			}
		}

		/// <summary>Gets the magnitude squared of the vector.</summary>
		public double MagnitudeSquared {
			get {
				return x * x + y * y + z * z;
			}
		}

		#endregion

		#region Static Properties

		/// <summary>Gets a vector with all components set to zero.</summary>
		public static Vector3 Zero { get { return new Vector3(0.0d); } }

		/// <summary>Gets a vector that points up.</summary>
		public static Vector3 Up { get { return new Vector3(0.0d, -1.0d, 0.0d); } }
		/// <summary>Gets a vector that points down.</summary>
		public static Vector3 Down { get { return new Vector3(0.0d, +1.0d, 0.0d); } }

		/// <summary>Gets a vector that points left.</summary>
		public static Vector3 Left { get { return new Vector3(-1.0d, 0.0d, 0.0d); } }
		/// <summary>Gets a vector that points right.</summary>
		public static Vector3 Right { get { return new Vector3(+1.0d, 0.0d, 0.0d); } }

		/// <summary>Gets a vector that points forwards.</summary>
		public static Vector3 Forward { get { return new Vector3(0.0d, 0.0d, +1.0d); } }
		/// <summary>Gets a vector that points backwards.</summary>
		public static Vector3 Backward { get { return new Vector3(0.0d, 0.0d, -1.0d); } }

		#endregion

		#region Constructors

		/// <summary>
		/// Creates an instance of a vector with the three components set to specified values.
		/// </summary>
		/// <param name="x">The value of the vector's X component.</param>
		/// <param name="y">The value of the vector's Y component.</param>
		/// <param name="z">The value of the vector's Z component.</param>
		public Vector3(double x, double y, double z) {
			this.x = x; this.y = y; this.z = z;
		}

		/// <summary>
		/// Creates an instance of a vector with the three components set to the same value.
		/// </summary>
		/// <param name="v">The value to set all three components to.</param>
		public Vector3(double v) {
			this.x = v; this.y = v; this.z = v;
		}

		#endregion

		#region General Methods

		/// <summary>Returns a formatted string describing the vector's components.</summary>
		public override string ToString() {
			return string.Format("'{0:N1} {1:N1} {2:N1}'", this.X, this.Y, this.Z);
		}

		/// <summary>Gets the hash code for this vector.</summary>
		public override int GetHashCode() {
			return this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Z.GetHashCode();
		}

		#endregion

		#region Arithmetic Methods

		/// <summary>
		/// Normalises a vector.
		/// </summary>
		/// <param name="v">The vector to normalise.</param>
		/// <returns>A vector of unit length.</returns>
		public static Vector3 Normalise(Vector3 v) {
			double Multiplier = 1f / v.Magnitude;
			return new Vector3() {
				X = v.x * Multiplier,
				Y = v.y * Multiplier,
				Z = v.z * Multiplier,
			};
		}

		/// <summary>Reflect a vector against a plane's surface normal.</summary>
		/// <param name="incident">The incident vector to reflect.</param>
		/// <param name="normal">The normal of the plane to reflect against.</param>
		public static Vector3 Reflect(Vector3 incident, Vector3 normal) {
			return incident - 2 * Vector3.Dot(incident, normal) * normal;
		}

		/// <summary>Calculates the dot (scalar) product of two vectors.</summary>
		public static double Dot(Vector3 a, Vector3 b) {
			return a.X * b.X + a.Y * b.Y + a.Z * b.Z;
		}

		/// <summary>Calculates the cross (vector) product of two vectors.</summary>
		public static Vector3 Cross(Vector3 a, Vector3 b) {
			return new Vector3(a.Y * b.Z - b.Y * a.Z, a.Z * b.X - b.Z * a.X, a.X * b.Y - b.X * a.Y);
		}

		/// <summary>Calculates the magnitude of each of the vector's components.</summary>
		public static Vector3 Abs(Vector3 v) {
			return new Vector3(Math.Abs(v.x), Math.Abs(v.y), Math.Abs(v.z));
		}

		/// <summary>Returns a vector made up of the smallest components of two vectors.</summary>
		public static Vector3 Min(Vector3 a, Vector3 b) {
			return new Vector3(Math.Min(a.X, b.X), Math.Min(a.Y, b.Y), Math.Min(a.Z, b.Z));
		}


		/// <summary>Returns a vector made up of the largest components of two vectors.</summary>
		public static Vector3 Max(Vector3 a, Vector3 b) {
			return new Vector3(Math.Max(a.X, b.X), Math.Max(a.Y, b.Y), Math.Max(a.Z, b.Z));
		}

		public static double DistanceSquared(Vector3 a, Vector3 b) {
			return (a - b).MagnitudeSquared;
		}

		public static Vector3 Refract(Vector3 incident, Vector3 normal, double refractiveIndexRatio) {
			
			double Cosθ1 = Vector3.Dot(normal, -incident);
			double Cosθ2 = 1.0d - refractiveIndexRatio * refractiveIndexRatio * (1.0d - Cosθ1 * Cosθ1);
			
			if (Cosθ2 < 0.0d) {
				//return Vector3.Reflect(incident, normal);
				return Vector3.Zero;
			} else {
				Cosθ2 = Math.Sqrt(Cosθ2);
				if (Cosθ1 > 0.0d) {
					return Vector3.Normalise((refractiveIndexRatio * incident) + (refractiveIndexRatio * Cosθ1 - Cosθ2) * normal);
				} else {
					return Vector3.Normalise((refractiveIndexRatio * incident) - (refractiveIndexRatio * Cosθ1 + Cosθ2) * normal);
				}
			}

		}

		/// <summary>
		/// Calculate the angle between two vectors.
		/// </summary>
		/// <returns>The angle between the two vectors.</returns>
		public static double GetAngle(Vector3 a, Vector3 b) {
			return Math.Acos(Vector3.Dot(a, b) / (a.Magnitude * b.Magnitude));
		}

		#endregion

		#region Operators

		/// <summary>Adds the components of two vectors together.</summary>
		public static Vector3 operator +(Vector3 a, Vector3 b) { return new Vector3(a.X + b.X, a.Y + b.Y, a.Z + b.Z); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector3 operator +(Vector3 a, double b) { return a + new Vector3(b); }

		/// <summary>Adds a constant value to the components of a vector.</summary>
		public static Vector3 operator +(double a, Vector3 b) { return new Vector3(a) + b; }

		/// <summary>Subtracts the components of one vector from another.</summary>
		public static Vector3 operator -(Vector3 a, Vector3 b) { return new Vector3(a.X - b.X, a.Y - b.Y, a.Z - b.Z); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector3 operator -(Vector3 a, double b) { return a - new Vector3(b); }

		/// <summary>Subtracts a constant from the components of a vector.</summary>
		public static Vector3 operator -(double a, Vector3 b) { return new Vector3(a) - b; }

		/// <summary>Negates the components of a vector.</summary>
		public static Vector3 operator -(Vector3 a) { return new Vector3(-a.X, -a.Y, -a.Z); }

		/// <summary>Multiplies the components of one vector with another.</summary>
		public static Vector3 operator *(Vector3 a, Vector3 b) { return new Vector3(a.X * b.X, a.Y * b.Y, a.Z * b.Z); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector3 operator *(Vector3 a, double b) { return a * new Vector3(b); }

		/// <summary>Multiplies the components of a vector with a constant value.</summary>
		public static Vector3 operator *(double a, Vector3 b) { return new Vector3(a) * b; }

		/// <summary>Divides the components of one vector by the components of another.</summary>
		public static Vector3 operator /(Vector3 a, Vector3 b) { return new Vector3(a.X / b.X, a.Y / b.Y, a.Z / b.Z); }

		/// <summary>Divides the components of a vector by a constant value.</summary>
		public static Vector3 operator /(Vector3 a, double b) { return a / new Vector3(b); }

		/// <summary>Divides a vector with all components set to a single value by another vector.</summary>
		public static Vector3 operator /(double a, Vector3 b) { return new Vector3(a) / b; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public static bool operator ==(Vector3 a, Vector3 b) { return a.X == b.X && a.Y == b.Y && a.Z == b.Z; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have differing components, false if both are the same.</returns>
		public static bool operator !=(Vector3 a, Vector3 b) { return a.X != b.X || a.Y != b.Y || a.Z != b.Z; }

		/// <summary>Compares the value of the components of two vectors.</summary>
		/// <returns>True if both vectors have the same components, false otherwise.</returns>
		public override bool Equals(object obj) {
			if (obj.GetType() != this.GetType()) return false;
			return (Vector3)obj == this;
		}

		#endregion
	}
}
